/*
 * Copyright (C) 2012 Robert Ancell.
 * Author: Robert Ancell <robert.ancell@canonical.com>
 * 
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version. See http://www.gnu.org/copyleft/gpl.html the full text of the
 * license.
 */

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "seat-weston.h"
#include "configuration.h"
#include "xserver-local.h"
#include "xsession.h"
#include "wayland-client.h" // FIXME: See https://bugs.freedesktop.org/show_bug.cgi?id=51740
#include "weston-protocol.h"

struct SeatWestonPrivate
{
    /* Compositor process */
    Process *compositor_process;

    /* Socket to communicate with compositor */
    int compositor_socket;
};

G_DEFINE_TYPE (SeatWeston, seat_weston, SEAT_TYPE);

static void
seat_weston_setup (Seat *seat)
{
    seat_set_can_switch (seat, TRUE);
    SEAT_CLASS (seat_weston_parent_class)->setup (seat);
}

static void
compositor_stopped_cb (Process *process, SeatWeston *seat)
{
    g_debug ("Stopping Weston seat, compositor terminated");
    seat_stop (SEAT (seat));
}

static gboolean
seat_weston_start (Seat *seat)
{
    gchar *command;
    int socket_fds[2], child_fd;
    gboolean result;

    if (socketpair (AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, socket_fds) == -1) 
    {
        //perror ("Failed to create compositor sockets");
        return FALSE;
    }
    child_fd = dup (socket_fds[0]);
    if (child_fd < 0)
        return FALSE;
    close (socket_fds[0]);
    SEAT_WESTON (seat)->priv->compositor_socket = socket_fds[1];

    if (socketpair (AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, socket_fds) == -1) 
    {
        //perror ("Failed to create compositor sockets");
        return FALSE;
    }

    command = g_strdup_printf ("weston --shell=system-compositor.so --display-manager-fd=%d", child_fd);
    process_set_command (SEAT_WESTON (seat)->priv->compositor_process, command);
    g_signal_connect (SEAT_WESTON (seat)->priv->compositor_process, "stopped", G_CALLBACK (compositor_stopped_cb), seat);
    result = process_start (SEAT_WESTON (seat)->priv->compositor_process);
    close (child_fd);

    if (!result)
        return FALSE;

    return SEAT_CLASS (seat_weston_parent_class)->start (seat);  
}

static DisplayServer *
seat_weston_create_display_server (Seat *seat)
{
    XServerLocal *xserver;
    const gchar *command = NULL, *layout = NULL, *config_file = NULL, *xdmcp_manager = NULL, *key_name = NULL;
    gboolean allow_tcp;
    gint port = 0;
    int socket_fds[2], child_fd;

    g_debug ("Starting X server on weston compositor");

    xserver = xserver_local_new ();

    if (socketpair (AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, socket_fds) == -1) 
    {
        //perror ("Failed to create compositor sockets");
        return FALSE;
    }
    child_fd = dup (socket_fds[0]);
    if (child_fd < 0)
        return FALSE;
    close (socket_fds[0]);
    xserver_local_set_wayland_socket (xserver, child_fd);

    command = seat_get_string_property (seat, "xserver-command");
    if (command)
        xserver_local_set_command (xserver, command);

    layout = seat_get_string_property (seat, "xserver-layout");
    if (layout)
        xserver_local_set_layout (xserver, layout);

    config_file = seat_get_string_property (seat, "xserver-config");
    if (config_file)
        xserver_local_set_config (xserver, config_file);
  
    allow_tcp = seat_get_boolean_property (seat, "xserver-allow-tcp");
    xserver_local_set_allow_tcp (xserver, allow_tcp);    

    xdmcp_manager = seat_get_string_property (seat, "xdmcp-manager");
    if (xdmcp_manager)
        xserver_local_set_xdmcp_server (xserver, xdmcp_manager);

    port = seat_get_integer_property (seat, "xdmcp-port");
    if (port > 0)
        xserver_local_set_xdmcp_port (xserver, port);

    key_name = seat_get_string_property (seat, "xdmcp-key");
    if (key_name)
    {
        gchar *dir, *path;
        GKeyFile *keys;
        gboolean result;
        GError *error = NULL;

        dir = config_get_string (config_get_instance (), "LightDM", "config-directory");
        path = g_build_filename (dir, "keys.conf", NULL);
        g_free (dir);

        keys = g_key_file_new ();
        result = g_key_file_load_from_file (keys, path, G_KEY_FILE_NONE, &error);
        if (error)
            g_debug ("Error getting key %s", error->message);
        g_clear_error (&error);      

        if (result)
        {
            gchar *key = NULL;

            if (g_key_file_has_key (keys, "keyring", key_name, NULL))
                key = g_key_file_get_string (keys, "keyring", key_name, NULL);
            else
                g_debug ("Key %s not defined", key_name);

            if (key)
                xserver_local_set_xdmcp_key (xserver, key);
            g_free (key);
        }

        g_free (path);
        g_key_file_free (keys);
    }

    return DISPLAY_SERVER (xserver);
}

static Session *
seat_weston_create_session (Seat *seat, Display *display)
{
    XServerLocal *xserver;
    XSession *session;
    gchar *tty;

    xserver = XSERVER_LOCAL (display_get_display_server (display));

    session = xsession_new (XSERVER (xserver));
    tty = g_strdup_printf ("/dev/tty%d", xserver_local_get_vt (xserver));
    session_set_tty (SESSION (session), tty);
    g_free (tty);

    return SESSION (session);
}

static void
seat_weston_set_active_display (Seat *seat, Display *display)
{
    // FIXME

    SEAT_CLASS (seat_weston_parent_class)->set_active_display (seat, display);
}

static void
seat_weston_run_script (Seat *seat, Display *display, Process *script)
{
    gchar *path;
    XServerLocal *xserver;

    xserver = XSERVER_LOCAL (display_get_display_server (display));
    path = xserver_local_get_authority_file_path (xserver);
    process_set_env (script, "DISPLAY", xserver_get_address (XSERVER (xserver)));
    process_set_env (script, "XAUTHORITY", path);
    g_free (path);

    SEAT_CLASS (seat_weston_parent_class)->run_script (seat, display, script);
}

static void
seat_weston_display_removed (Seat *seat, Display *display)
{
    if (seat_get_is_stopping (seat))
        return;

    /* If this is the only display and it failed to start then stop this seat */
    if (g_list_length (seat_get_displays (seat)) == 0 && !display_get_is_ready (display))
    {
        g_debug ("Stopping Weston seat, failed to start a display");
        //seat_stop (seat);
        return;
    }

    /* Show a new greeter */  
    if (display == seat_get_active_display (seat))
    {
        g_debug ("Active display stopped, switching to greeter");
        seat_switch_to_greeter (seat);
    }
}

static void
seat_weston_init (SeatWeston *seat)
{
    seat->priv = G_TYPE_INSTANCE_GET_PRIVATE (seat, SEAT_WESTON_TYPE, SeatWestonPrivate);
    seat->priv->compositor_process = process_new ();
}

static void
seat_weston_finalize (GObject *object)
{
    SeatWeston *seat = SEAT_WESTON (object);

    g_object_unref (seat->priv->compositor_process);

    G_OBJECT_CLASS (seat_weston_parent_class)->finalize (object);
}

static void
seat_weston_class_init (SeatWestonClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);  
    SeatClass *seat_class = SEAT_CLASS (klass);

    object_class->finalize = seat_weston_finalize;
    seat_class->setup = seat_weston_setup;
    seat_class->start = seat_weston_start;
    seat_class->create_display_server = seat_weston_create_display_server;
    seat_class->create_session = seat_weston_create_session;
    seat_class->set_active_display = seat_weston_set_active_display;
    seat_class->run_script = seat_weston_run_script;
    seat_class->display_removed = seat_weston_display_removed;

    g_type_class_add_private (klass, sizeof (SeatWestonPrivate));
}
